@@ -229,6 +229,415 @@ class Binarizer:
 
             json.dump(previous_content, jfile, indent=4)
 
+# class PatchTransform3D:
+#     """
+#     Operation used to apply a custom transformation patch-wise on some input dataset.
+#     """
+#
+#     __name__ = 'PatchTransform3D'
+#     def __init__(self,trace,patch_transform_function):
+#         """
+#         Initialize the operation. The PatchTransform3D operation is used to apply a transformation function specified
+#         by the user patch-wise on some input dataset.
+#
+#         :param trace: (bmmltools.core.tracer.Trace object) trace on which the operation act.
+#         :param patch_transform_function: (function) function used to extract information from the patch. It has to be a
+#                                          single input function. the only input have to be a patch of the dataset on
+#                                          which this operation is applied.
+#         """
+#         self.trace = trace
+#         self.output_dataset = ['post_'+get_capitals(PatchTransform3D.__name__)+'_dataset']
+#         self.pt = ParametersTracker(self)
+#         self._patch_transform = patch_transform_function
+#
+#     def i(self,x):
+#         """
+#         Specify operation inputs.
+#
+#         :param x: (str, list[str]) name of the input dataset used in this operation. If more than one input is required
+#                   a list with the name of the input dataset need to be specified following the specific order described
+#                   in the bmmltools documentation.
+#         """
+#         if type(x) is str:
+#
+#             self.input_dataset = [x]
+#
+#         elif type(x) in [tuple,list]:
+#
+#             self.input_dataset = x
+#
+#         else:
+#
+#             raise TypeError
+#
+#         return self
+#
+#     def o(self,x):
+#         """
+#         Specify operation outputs.
+#
+#         :param x: (str, list[str]) optional, name or list of names of the operations output. If more than one output
+#                   is present, in the list one needs to specify all the names of the output. If nothing is specified,
+#                   the default name is used. The default name is
+#
+#                                 'post_[Capital Letters + Numbers in the operation class name]_dataset'.
+#
+#                   If more than one output is present, some additional words or numbers are added to the default name
+#                   in order to differentiate the different outputs.
+#         """
+#         if type(x) is str:
+#
+#             self.output_dataset = [x]
+#
+#         elif type(x) in [tuple, list]:
+#
+#             self.output_dataset = x
+#
+#         else:
+#
+#             raise TypeError
+#
+#         return self
+#
+#     def io(self,x,y):
+#         """
+#         Specify operation inputs and outputs.
+#
+#         :param x: (str, list[str]) name of the input dataset used in this operation. If more than one input is required
+#                   a list with the name of the input dataset need to be specified following the specific order described
+#                   in the bmmltools documentation.
+#         :param y: (str, list[str]) optional, name or list of names of the operations output. If more than one output
+#                   is present, in the list one needs to specify all the names of the output. If nothing is specified,
+#                   the default name is used. The default name is
+#
+#                                 'post_[Capital Letters + Numbers in the operation class name]_dataset'.
+#
+#                   If more than one output is present, some additional words or numbers are added to the default name
+#                   in order to differentiate the different outputs.
+#         """
+#         self.i(x)
+#         self.o(y)
+#         return self
+#
+#     def apply(self,patch_shape,transform_name = None,random_patches=False,n_random_patches=4000):
+#         """
+#         Apply the operation to the trace. The function 'transform_function' specified during the operation
+#         initialization is applied to various patches of the input data. The application can be done:
+#
+#             * on the ZYX-grid, where the transformation for all the patches obtained dividing the sample into an regular
+#             ZYX grid which is constructed from the patch chosen;
+#
+#             * at random, where patches are take in random position of the sample.
+#
+#         :param patch_shape: (tuple[int]) shape of the patch used in this operation.
+#         :param transform_name: (str) optional, name given to the transformation output (one of the two keys of the
+#                                output dictionary of this operation)
+#         :param random_patches: (bool) optional, if True the patch transform is applied in random points of the input
+#                                dataset, otherwise an ZYX-grid is used.
+#         :param n_random_patches: (int) optional, number of random patches taken (this variable is considered only when
+#                                  random patches are used).
+#         """
+#         #
+#         with self.pt:
+#
+#             self.patch_shape = patch_shape
+#             self.transform_name = transform_name
+#             self.random_patches = random_patches
+#             self.n_random_patches = n_random_patches
+#
+#         #
+#         if self.transform_name is None:
+#
+#             self.transform_name = 'transformed_patch'
+#
+#         with h5py.File(self.trace.hdf5_trace_path, 'r') as file:
+#
+#             self.data_shape = file[self.trace._hdf5_inner_path+self.input_dataset[0]].shape
+#
+#         self.patch_space_shape = tuple(np.array(self.data_shape) // np.array(self.patch_shape))
+#
+#         print('>----------------<')
+#         print('Patch Transform 3d')
+#         print('({})'.format('using random patches' if self.random_patches else 'using zyx-grid'))
+#         if self.random_patches:
+#
+#             self._random_patches_transform(self.transform_name)
+#
+#         else:
+#
+#             self._on_grid_patches_transform(self.transform_name)
+#
+#         print('>----------------<',end='\r')
+#         self._update_graph_dict()
+#         self._update_params_dict()
+#         if self.trace.enable_operations_tracking:
+#
+#             self.trace.add_operation(PatchTransform3D.__name__,self,self.pt.get_current_tracked_params())
+#
+#         return self.output_dataset
+#
+#     def _on_grid_patches_transform(self,name):
+#         """
+#         Compute the patch transform using a regular ZYX-grid.
+#
+#         :param name: (str) name of the dataset (dictionary key in this case) where the transformed patches are saved.
+#         """
+#         # get coordinates
+#         Zs, Ys, Xs = np.meshgrid(np.arange(0,self.patch_space_shape[0]),
+#                                  np.arange(0,self.patch_space_shape[1]),
+#                                  np.arange(0,self.patch_space_shape[2]))
+#         ZYXs = np.vstack([Zs.flatten(), Ys.flatten(), Xs.flatten()]).transpose()
+#         n_patches = len(ZYXs)
+#
+#         # get necessary trace attributes before the hdf5 file is open
+#         hdf5_inner_path = self.trace._hdf5_inner_path
+#         group_name = self.trace.group_name
+#         traced_dict = self.trace.dict_variables_traced_on_hd
+#         traced_pandas = self.trace.pandas_variables_traced_on_hd
+#         traced_numpy = self.trace.numpy_variables_traced_on_hd
+#
+#         #
+#         with h5py.File(self.trace.hdf5_trace_path,'a') as file:
+#
+#             # delete dataset with the same name of the output if present
+#             try:
+#
+#                 del file[self.trace._hdf5_inner_path+self.output_dataset[0]]
+#                 traced_dict = [elem for elem in traced_dict
+#                                if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#                 traced_numpy = [elem for elem in traced_numpy
+#                                 if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#                 traced_pandas = [elem for elem in traced_pandas
+#                                  if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#
+#             except:
+#
+#                 pass
+#
+#             group_db = file.create_group(name=hdf5_inner_path + self.output_dataset[0])
+#             tmp_db = group_db.create_dataset('{}_tmp'.format(name),
+#                                              shape=(n_patches,)+self.patch_shape,
+#                                              dtype=np.float64,
+#                                              chunks=(1,)+self.patch_shape)
+#             in_db = file[hdf5_inner_path+self.input_dataset[0]]
+#             nonnull_Zs = []
+#             nonnull_Ys = []
+#             nonnull_Xs = []
+#             m = 0
+#             for n in range(n_patches):
+#
+#                 # select patch
+#                 z,y,x = ZYXs[n]
+#                 zmin = int(z*self.patch_shape[0])
+#                 zmax = zmin+self.patch_shape[0]
+#                 ymin = int(y*self.patch_shape[1])
+#                 ymax = ymin+self.patch_shape[1]
+#                 xmin = int(x*self.patch_shape[2])
+#                 xmax = xmin+self.patch_shape[2]
+#                 patch = in_db[zmin:zmax,ymin:ymax,xmin:xmax]
+#                 if np.sum(patch) > 0:
+#
+#                     # compute transformed patch
+#                     tmp_db[m,:,:,:] = self._patch_transform(patch)
+#                     nonnull_Zs.append(z)
+#                     nonnull_Ys.append(y)
+#                     nonnull_Xs.append(x)
+#                     m += 1
+#
+#                 print('patch {}/{}'.format(n+1,n_patches), end='\r')
+#
+#             group_db.create_dataset(name,data=tmp_db[:len(nonnull_Zs),:,:,:],
+#                                          dtype=np.float64,
+#                                          chunks=(1,) + self.patch_shape)
+#             del group_db['{}_tmp'.format(name)]
+#
+#             # update trace attributes
+#             traced_dict.append('{}.{}'.format(group_name, self.output_dataset[0]))
+#             file[hdf5_inner_path].attrs['dict_variables_traced_on_hd'] = list(set(traced_dict))
+#
+#             traced_numpy.append('{}.{}.{}'.format(group_name, self.output_dataset[0],name))
+#             file[hdf5_inner_path].attrs['numpy_variables_traced_on_hd'] = \
+#                 list(set([el for el in traced_numpy if group_name in el]))
+#
+#         patch_space_coords_df = pd.DataFrame({'Z': nonnull_Zs,
+#                                               'Y': nonnull_Ys,
+#                                               'X': nonnull_Xs})
+#         patch_space_coords_df.to_hdf(self.trace.hdf5_trace_path,
+#                                      hdf5_inner_path+self.output_dataset[0]+'/patch_space_coordinates',
+#                                      mode='a', format='t', data_columns=True)
+#         with h5py.File(self.trace.hdf5_trace_path, 'a') as file:
+#
+#             traced_pandas.append('{}.{}.{}'.format(group_name,self.output_dataset[0],'patch_space_coordinates'))
+#             file[hdf5_inner_path].attrs['pandas_variables_traced_on_hd'] = \
+#                 list(set([el for el in traced_pandas if group_name in el]))
+#
+#         self.trace.numpy_variables_traced_on_hd = list(set(traced_numpy))
+#         self.trace.pandas_variables_traced_on_hd = list(set(traced_pandas))
+#         self.trace.dict_variables_traced_on_hd = list(set(traced_dict))
+#
+#     def _random_patches_transform(self,name):
+#         """
+#         Compute the patch transform using random patches.
+#
+#         :param name: (str) name of the dataset (dictionary key in this case) where the transformed patches are saved.
+#         """
+#         # set seed if present
+#         if hasattr(self.trace,'seed'):
+#
+#             np.random.seed(self.trace.seed)
+#
+#         # get necessary trace attributes before the hdf5 file is open
+#         hdf5_inner_path = self.trace._hdf5_inner_path
+#         group_name = self.trace.group_name
+#         traced_dict = self.trace.dict_variables_traced_on_hd
+#         traced_pandas = self.trace.pandas_variables_traced_on_hd
+#         traced_numpy = self.trace.numpy_variables_traced_on_hd
+#
+#         #
+#         with h5py.File(self.trace.hdf5_trace_path,'a') as file:
+#
+#             # try:
+#
+#             if self.output_dataset[0] in file[hdf5_inner_path].keys():
+#
+#                 del file[self.trace._hdf5_inner_path+self.output_dataset[0]]
+#                 traced_dict = [elem for elem in traced_dict
+#                                if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#                 traced_numpy = [elem for elem in traced_numpy
+#                                 if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#                 traced_pandas = [elem for elem in traced_pandas
+#                                  if '{}.{}'.format(group_name,self.output_dataset[0]) not in elem]
+#
+#             #
+#             group_db = file.create_group(name=hdf5_inner_path + self.output_dataset[0])
+#             db = group_db.create_dataset('{}_tmp'.format(name),
+#                                     shape=(self.n_random_patches,)+self.patch_shape,
+#                                     dtype=np.float64,
+#                                     chunks=(1,)+self.patch_shape)
+#             in_db = file[hdf5_inner_path+self.input_dataset[0]]
+#             Zs = []
+#             Ys = []
+#             Xs = []
+#             while True:
+#
+#                 zmin = np.random.randint(self.patch_shape[0],self.data_shape[0]-self.patch_shape[0])
+#                 zmax = zmin+self.patch_shape[0]
+#                 ymin = np.random.randint(self.patch_shape[1],self.data_shape[1]-self.patch_shape[1])
+#                 ymax = ymin+self.patch_shape[1]
+#                 xmin = np.random.randint(self.patch_shape[2],self.data_shape[2]-self.patch_shape[2])
+#                 xmax = xmin+self.patch_shape[2]
+#                 patch = in_db[zmin:zmax,ymin:ymax,xmin:xmax]
+#                 if np.sum(patch) > 0:
+#
+#                     # compute transformed patch
+#                     db[len(Zs),:,:,:] = self._patch_transform(patch)
+#                     Zs.append(zmin/self.patch_shape[0])
+#                     Ys.append(ymin/self.patch_shape[1])
+#                     Xs.append(xmin/self.patch_shape[2])
+#
+#                     print('patch: {}/{}'.format(len(Zs),self.n_random_patches),end='\r')
+#                     if len(Zs) == self.n_random_patches:
+#
+#                         break
+#
+#             # update trace attributes
+#             traced_dict.append('{}.{}'.format(group_name, self.output_dataset[0]))
+#             file[hdf5_inner_path].attrs['dict_variables_traced_on_hd'] = list(set(traced_dict))
+#
+#             traced_numpy.append('{}.{}.{}'.format(group_name, self.output_dataset[0],name))
+#             file[hdf5_inner_path].attrs['numpy_variables_traced_on_hd'] = \
+#                 list(set([el for el in traced_numpy if group_name in el]))
+#
+#         patch_space_coords_df = pd.DataFrame({'Z': Zs,'Y': Ys,'X': Xs})
+#         patch_space_coords_df.to_hdf(self.trace.hdf5_trace_path,
+#                                      hdf5_inner_path + self.output_dataset[0] + '/patch_space_coordinates',
+#                                      mode='a', format='t', data_columns=True)
+#         with h5py.File(self.trace.hdf5_trace_path, 'a') as file:
+#
+#             traced_pandas.append(
+#                 '{}.{}.{}'.format(group_name, self.output_dataset[0], 'patch_space_coordinates'))
+#             file[hdf5_inner_path].attrs['pandas_variables_traced_on_hd'] = \
+#                 list(set([el for el in traced_pandas if group_name in el]))
+#
+#         self.trace.numpy_variables_traced_on_hd = list(set(traced_numpy))
+#         self.trace.pandas_variables_traced_on_hd = list(set(traced_pandas))
+#         self.trace.dict_variables_traced_on_hd = list(set(traced_dict))
+#
+#     def read(self):
+#         """
+#         Read operation results in the hdf5 file. For this plugin, the results are organized in the form of a pandas
+#         dataframe (patch coordinates) and numpy array (transformation result). The first is saved as csv file, while the
+#         second as npy file.
+#         """
+#         trace_pos = self.trace.get_trace_pos()
+#         path_to_readings_folder = manage_path(
+#             self.trace.trace_readings_path() + os.sep + PatchTransform3D.__name__+'_{}'.format(trace_pos))
+#         res = self.trace.__getattribute__(self.output_dataset[0])
+#         for key in res.keys():
+#
+#             if type(res[key]) == pd.DataFrame:
+#
+#                 res[key].to_csv(path_to_readings_folder + os.sep + key + '.csv')
+#
+#             else:
+#
+#                 np.save(path_to_readings_folder + os.sep + key + '.npy', res[key])
+#
+#     def _update_graph_dict(self):
+#         """
+#         Update the graph dictionary in the trace json adding the inputs and outputs of this operation.
+#         """
+#         if self.trace.enable_trace_graph == True:
+#
+#             tmp_dict = self.trace.trace_graph_dict
+#             link_list = []
+#             for i_name in self.input_dataset:
+#
+#                 for o_name in self.output_dataset:
+#
+#                     link_list.append((i_name,o_name))
+#
+#             key = len(tmp_dict)
+#             if tmp_dict[key-1]['node'] == self.output_dataset:
+#
+#                 key = key-1
+#
+#             tmp_dict.update({key:{'node': self.output_dataset,
+#                                    'link': list(set(link_list)),
+#                                    'edge': PatchTransform3D.__name__}})
+#             self.trace.__standard_setattr__('trace_graph_dict',tmp_dict)
+#             with open(self.trace.json_trace_path, 'r') as jfile:
+#
+#                 previous_content = json.load(jfile)
+#
+#             previous_content[self.trace.group_name]['trace_graph'] = tmp_dict
+#             with open(self.trace.json_trace_path, 'w') as jfile:
+#
+#                 json.dump(previous_content, jfile, indent=4)
+#
+#     def _update_params_dict(self):
+#         """
+#         Update the parameter dictionary in the trace json adding parameters of this operation.
+#         """
+#         with open(self.trace.json_trace_path, 'r') as jfile:
+#
+#             previous_content = json.load(jfile)
+#
+#         params_dict = previous_content[self.trace.group_name]['ops_parameters']
+#         key = len(params_dict)
+#         if key > 0 and \
+#                 previous_content[self.trace.group_name]['trace_graph'][str(key-1)]['node'] == self.output_dataset:
+#
+#             key = key-1
+#             del params_dict[str(key)]
+#
+#         params_dict.update({key: {'op': PatchTransform3D.__name__,
+#                                   'p': self.pt.get_current_tracked_params()}})
+#         previous_content[self.trace.group_name]['ops_parameters'] = params_dict
+#         with open(self.trace.json_trace_path, 'w') as jfile:
+#
+#             json.dump(previous_content, jfile, indent=4)
+
 class PatchTransform3D:
     """
     Operation used to apply a custom transformation patch-wise on some input dataset.
@@ -319,7 +728,7 @@ class PatchTransform3D:
         self.o(y)
         return self
 
-    def apply(self,patch_shape,transform_name = None,random_patches=False,n_random_patches=4000):
+    def apply(self,patch_shape,stride=None,transform_name = None,random_patches=False,n_random_patches=4000):
         """
         Apply the operation to the trace. The function 'transform_function' specified during the operation
         initialization is applied to various patches of the input data. The application can be done:
@@ -329,7 +738,9 @@ class PatchTransform3D:
 
             * at random, where patches are take in random position of the sample.
 
-        :param patch_shape: (tuple[int]) shape of the patch used in this operation.
+        :param patch_shape: (tuple[int]) shape of the input patch used in this operation.
+        :param stride: (None or tuple[int]) optional, stride along each dimension (used when the transformation is
+                       taken along a ZYX-grid).
         :param transform_name: (str) optional, name given to the transformation output (one of the two keys of the
                                output dictionary of this operation)
         :param random_patches: (bool) optional, if True the patch transform is applied in random points of the input
@@ -341,6 +752,7 @@ class PatchTransform3D:
         with self.pt:
 
             self.patch_shape = patch_shape
+            self.stride = stride
             self.transform_name = transform_name
             self.random_patches = random_patches
             self.n_random_patches = n_random_patches
@@ -354,7 +766,11 @@ class PatchTransform3D:
 
             self.data_shape = file[self.trace._hdf5_inner_path+self.input_dataset[0]].shape
 
+        self.transformation_output_shape = self._patch_transform(np.zeros(self.patch_shape)).shape
         self.patch_space_shape = tuple(np.array(self.data_shape) // np.array(self.patch_shape))
+        if self.stride == None:
+
+            self.stride = self.patch_shape
 
         print('>----------------<')
         print('Patch Transform 3d')
@@ -383,10 +799,16 @@ class PatchTransform3D:
         :param name: (str) name of the dataset (dictionary key in this case) where the transformed patches are saved.
         """
         # get coordinates
-        Zs, Ys, Xs = np.meshgrid(np.arange(0,self.patch_space_shape[0]),
-                                 np.arange(0,self.patch_space_shape[1]),
-                                 np.arange(0,self.patch_space_shape[2]))
+        patch_space_dz = self.stride[0]/self.patch_shape[0]
+        patch_space_dy = self.stride[1]/self.patch_shape[1]
+        patch_space_dx = self.stride[2]/self.patch_shape[2]
+        Zs, Ys, Xs = np.meshgrid(np.arange(0,self.patch_space_shape[0],patch_space_dz),
+                                 np.arange(0,self.patch_space_shape[1],patch_space_dy),
+                                 np.arange(0,self.patch_space_shape[2],patch_space_dx))
         ZYXs = np.vstack([Zs.flatten(), Ys.flatten(), Xs.flatten()]).transpose()
+
+        self.APP = ZYXs
+
         n_patches = len(ZYXs)
 
         # get necessary trace attributes before the hdf5 file is open
@@ -416,9 +838,9 @@ class PatchTransform3D:
 
             group_db = file.create_group(name=hdf5_inner_path + self.output_dataset[0])
             tmp_db = group_db.create_dataset('{}_tmp'.format(name),
-                                             shape=(n_patches,)+self.patch_shape,
+                                             shape=(n_patches,)+self.transformation_output_shape,
                                              dtype=np.float64,
-                                             chunks=(1,)+self.patch_shape)
+                                             chunks=(1,)+self.transformation_output_shape)
             in_db = file[hdf5_inner_path+self.input_dataset[0]]
             nonnull_Zs = []
             nonnull_Ys = []
@@ -437,7 +859,7 @@ class PatchTransform3D:
                 patch = in_db[zmin:zmax,ymin:ymax,xmin:xmax]
                 if np.sum(patch) > 0:
 
-                    # compute transformed patch
+                    # compute transformed patch and coordinates in the output patch shape
                     tmp_db[m,:,:,:] = self._patch_transform(patch)
                     nonnull_Zs.append(z)
                     nonnull_Ys.append(y)
@@ -448,7 +870,7 @@ class PatchTransform3D:
 
             group_db.create_dataset(name,data=tmp_db[:len(nonnull_Zs),:,:,:],
                                          dtype=np.float64,
-                                         chunks=(1,) + self.patch_shape)
+                                         chunks=(1,) + self.transformation_output_shape)
             del group_db['{}_tmp'.format(name)]
 
             # update trace attributes
@@ -510,10 +932,10 @@ class PatchTransform3D:
 
             #
             group_db = file.create_group(name=hdf5_inner_path + self.output_dataset[0])
-            db = group_db.create_dataset('{}_tmp'.format(name),
-                                    shape=(self.n_random_patches,)+self.patch_shape,
+            db = group_db.create_dataset('{}'.format(name),
+                                    shape=(self.n_random_patches,)+self.transformation_output_shape,
                                     dtype=np.float64,
-                                    chunks=(1,)+self.patch_shape)
+                                    chunks=(1,)+self.transformation_output_shape)
             in_db = file[hdf5_inner_path+self.input_dataset[0]]
             Zs = []
             Ys = []
@@ -1784,4 +2206,4 @@ class DataStandardization:
         previous_content[self.trace.group_name]['ops_parameters'] = params_dict
         with open(self.trace.json_trace_path, 'w') as jfile:
 
-            json.dump(previous_content, jfile, indent=4)
+            json.dump(previous_content, jfile, indent=4)
\ No newline at end of file
